import os
import tempfile
import pandas as pd
import json
from singleton_tools import SingletonMeta


class PandasTools(metaclass=SingletonMeta):
    def __init__(self):
        pass

    # <editor-fold desc="Archivos temporales    ----------------------------------------------------------------------------------------------------------------------------------">
    @staticmethod
    def save_temporary_df(df: pd.DataFrame) -> tempfile.NamedTemporaryFile:
        """
        Guarda un DataFrame en un archivo temporal y devuelve la ruta del archivo.
        :param df:DataFrame de pandas a guardar.
        :return:Un objeto de archivo temporal que necesita ser cerrado después de su uso.
        """
        # Crear un archivo temporal para el DataFrame
        temp_file = tempfile.NamedTemporaryFile(mode='w+', delete=False, suffix='.pkl')

        # Guardar el DataFrame en el archivo temporal
        df.to_pickle(temp_file.name)

        # Devolver el objeto de archivo temporal para su posterior uso
        return temp_file

    @staticmethod
    def load_temporary_df(temp_file_object) -> pd.DataFrame:
        """
        Lee un DataFrame desde un archivo temporal dado.
        :param temp_file_object: Objeto de archivo temporal desde el cual leer el DataFrame.
        :return: DataFrame de pandas leído desde el archivo.
        """

        # Leer el DataFrame desde el archivo temporal
        df = pd.read_pickle(temp_file_object.name)

        # Cerrar y eliminar el archivo temporal
        temp_file_object.close()

        return df

    # </editor-fold>

    # <editor-fold desc="Json Methods    -----------------------------------------------------------------------------------------------------------------------------------------">

    @staticmethod
    def save_json_from_dict(path_and_file_with_extension: str, to_save_dict_object: dict):
        with open(path_and_file_with_extension, 'w', encoding='utf-8') as archivo_json:
            json.dump(to_save_dict_object, archivo_json, ensure_ascii=False, indent=4)

    @staticmethod
    def load_json_in_dict(path_and_file_with_extension: str) -> dict:
        with open(path_and_file_with_extension, 'r', encoding='utf-8') as f:
            return json.load(f)

    # </editor-fold>

    # <editor-fold desc="Utilidades varias    ------------------------------------------------------------------------------------------------------------------------------------">

    @staticmethod
    def boolean_imputer(df: pd.DataFrame, colname: str, value_to_fill):
        """
        Este método se usa para imputar columnas boleanas transformandolas a entero y agregando un nuevo valor para los NA
        :param df:
        :param colname:
        :param value_to_fill:
        :return:
        """
        df[colname] = pd.to_numeric(df[colname], errors='coerce').astype('Int64')
        df[colname] = df[colname].fillna(value_to_fill)

    @staticmethod
    def stringColsToList(str_text: str):
        lst = list(str_text.replace("[", "").replace("]", "").split(sep=","))
        return [z.strip() for z in lst]

    # </editor-fold>

    # <editor-fold desc="Transformación y tipado    ------------------------------------------------------------------------------------------------------------------------------">

    def tipping_by_dict(self, df: pd.DataFrame, colname_type_dict: dict, date_format: str) -> pd.DataFrame:
        """
        Tipar las columnas que se encuentran en un dict. Options: [Int64, float64, datetime, object, category]
        :param date_format:
        :param df:
        :param colname_type_dict:
        :return:
        """
        df_cols_list: list = [z for z in df.columns]
        for colname, coltype in colname_type_dict.items():
            if colname in df_cols_list:
                match coltype:
                    case "Int64":
                        df[colname] = self.to_pandas_int64(df[colname])
                    case "float64":
                        df[colname] = self.to_pandas_float64(df[colname])
                    case "datetime":
                        df[colname] = self.to_pandas_datetime(df[colname], date_format)
                    case "object":
                        df[colname] = self.to_pandas_object(df[colname])
                    case "category":
                        df[colname] = self.to_pandas_categorical(df[colname])
                    case _:
                        pass
        return df

    @staticmethod
    def to_pandas_int64(df_column: pd.Series):
        return pd.to_numeric(df_column, errors='coerce').astype('Int64')

    @staticmethod
    def to_pandas_bool(df_column: pd.Series, firstInt64: bool = True):
        if firstInt64:
            return pd.to_numeric(df_column, errors='coerce').astype('Int64').astype(pd.BooleanDtype())
        else:
            return df_column.astype(pd.BooleanDtype())

    @staticmethod
    def to_pandas_float64(df_column: pd.Series):
        return pd.to_numeric(df_column, errors='coerce').astype('float64')

    @staticmethod
    def to_pandas_datetime(df_column: pd.Series | str, formatter: str = "%Y-%m-%d"):
        try:
            return pd.to_datetime(df_column, format=f"{formatter}", errors="coerce")
        except TypeError:
            r = pd.to_datetime(df_column, format=formatter.replace(" %H:%M:%S", ""), errors="coerce")
            if r.isnull().sum() == r.size:
                raise TypeError("El formatter indicado ha arrojado un ValueError, se ha reintentado con %Y-%m-%d y ha hecho nula la columna")
            return r

    @staticmethod
    def to_pandas_timedelta(df_column: pd.Series | str):
        return pd.to_timedelta(df_column, errors="coerce")

    @staticmethod
    def to_pandas_object(df_column: pd.Series):
        return df_column.astype("object", copy=False)

    @staticmethod
    def to_pandas_string(df_column: pd.Series):
        return df_column.astype(pd.StringDtype())

    @staticmethod
    def to_pandas_categorical(df_column: pd.Series):
        return pd.Categorical(df_column)

    @staticmethod
    def to_pandas_categorical_ordered(df_column: pd.Series, ordered_values_list: list):
        return pd.Categorical(df_column, categories=ordered_values_list, ordered=True)

    # </editor-fold>

    # <editor-fold desc="Carga y guardado de dataframes    -----------------------------------------------------------------------------------------------------------------------">

    @staticmethod
    def read_csv_infering(file_path: str, sep: str | None = None, usecols: list | None = None) -> pd.DataFrame:
        """
        Metodo para leer archivos csv
        :param file_path: Ruta del archivo y extension (csv)
        :param sep: Default = "," -> Separador del csv
        :param usecols: En caso de querer solo algunas columnas, las especificamos en una lista
        :return: pd.DataFrame
        """
        return pd.read_csv(file_path, sep=sep, header=0, usecols=usecols, engine="python")

    @staticmethod
    def read_excel_infering(file_path: str, sheet_name: str | int = 0, usecols: list | None = None) -> pd.DataFrame:
        """
        Método para leer archivos Excel.
        :param file_path: Ruta del archivo y extensión (xlsx, xls).
        :param sheet_name: Nombre o índice de la hoja de cálculo que se desea leer. Por defecto es la primera hoja.
        :param usecols: En caso de querer solo algunas columnas, las especificamos en una lista.
        :return: pd.DataFrame
        """
        return pd.read_excel(file_path, sheet_name=sheet_name, usecols=usecols)

    @staticmethod
    def df_to_csv(path: str, df: pd.DataFrame, index: bool = False):
        """
        Metodo para guardar dataframes de pandas en archivos csv
        :param path: ruta/archivo.csv
        :param df: pd.Dataframe
        :param index: Añadir columna de indices, por defecto false
        :return: True si ha creado el archivo, False si ya existe
        """
        if not os.path.exists(path):
            print(f"---> dfToCsv: Guardando {path} ...")
            df.to_csv(path, index=index)
            print(f"---> dfToCsv: {path} guardado con exito. Su shape es: {df.shape}")
        else:
            print(f"---> dfToCsv: El archivo {path} ya existe")

    @staticmethod
    def df_to_parquet(path: str, df: pd.DataFrame, index: bool = False, replace_if_exists: bool = False) -> bool:
        """
        Metodo para guardar dataframes de pandas en archivos parquet (requiere pyarrow)

        :param path: Ruta completa del archivo. EXAMPLE: 'data/output/archivo.parquet'
        :param df: DataFrame de pandas a guardar
        :param index: Añadir columna de índices, por defecto False
        :param replace_if_exists: Machacar el archivo si ya existe, por defecto False
        :return: True si ha creado/reemplazado el archivo, False si ya existe y no se reemplaza
        """
        if not os.path.exists(path):
            print(f"---> dfToParquet: Guardando {path} ...")
            df.to_parquet(path, index=index)
            print(f"---> dfToParquet: {path} guardado con éxito. Su shape es: {df.shape}")
            return True
        else:
            if replace_if_exists:
                print(f"---> dfToParquet: El archivo {path} ya existe. Reemplazando...")
                os.remove(path)
                df.to_parquet(path, index=index)
                print(f"---> dfToParquet: {path} reemplazado con éxito. Su shape es: {df.shape}")
                return True
            else:
                print(f"---> dfToParquet: El archivo {path} ya existe. No se reemplaza.")
                return False

    @staticmethod
    def read_parquet_infering(file_path: str) -> pd.DataFrame:
        """
        Metodo para leer archivos Parquet (requiere pyarrow)
        :param file_path: Ruta del archivo y extension (Parquet)
        :return: pd.DataFrame
        """
        return pd.read_parquet(file_path)

    @staticmethod
    def df_to_json(path: str, df: pd.DataFrame, index: bool = False, dtype_dict: dict | None = None):
        """
        Metodo para guardar dataframes de pandas en archivos json
        :param dtype_dict: Diccionario de tipos de columna del json
        :param path: ruta/archivo.json
        :param df: pd.Dataframe
        :param index: Añadir columna de indices, por defecto false
        :return: None
        """
        if dtype_dict is not None:
            df.to_json(path, index=index, orient='records', lines=False, dtype=dtype_dict)
        else:
            df.to_json(path)

    # </editor-fold>



import pandas as pd
import polars as pl
import colorama
import warnings
from tabulate import tabulate
from colorama import Fore, Back, Style
from singleton_tools import SingletonMeta
colorama.init(autoreset=True)


class InfoTools(metaclass=SingletonMeta):
    def __init__(self, show_info: bool = True):
        self.show_info: bool = show_info

    @staticmethod
    def create_df_info(df: pl.DataFrame | pd.DataFrame, categoric_limit: int = 8, table_fmt: str = "fancy_grid"):
        is_polars = isinstance(df, pl.DataFrame)
        shape = df.shape

        print("#####################################################################################")
        print("Shape del DataFrame:")
        print(f"Filas: {shape[0]}")
        print(f"Columnas: {shape[1]}\n")

        info = []
        for col in df.columns:
            if is_polars:
                value_counts = df[col].value_counts(sort=True)
                unique_counts = dict(zip(
                    value_counts[col].to_list(),
                    value_counts["count"].to_list()
                ))
                null_count = df[col].null_count()
                dtype_str = str(df[col].dtype)
            else:
                unique_counts = df[col].value_counts().to_dict()
                null_count = df[col].isnull().sum()
                dtype_str = str(df[col].dtype)

            unique_counts_str = [f"{k}({v})" for k, v in unique_counts.items()]
            valores_unicos = ", ".join(unique_counts_str) if len(unique_counts) < categoric_limit else 'NO CATEGORICA'

            info.append({
                "NombreColumna": col,
                "NumeroValoresUnicos": len(unique_counts),
                "NumeroNulos": null_count,
                "ValoresUnicos": valores_unicos,
                "Dtype": dtype_str
            })

        # Crear DataFrame según el tipo de entrada
        if is_polars:
            info_df = pl.DataFrame(info)
            info_pandas = info_df.to_pandas()
        else:
            info_pandas = pd.DataFrame(info)

        print(tabulate(info_pandas, showindex=False, headers=info_pandas.columns, tablefmt=table_fmt))

    @staticmethod
    def print_polars_native(df: pl.DataFrame, row_print: int = 10):
        """
        Metodo alternativo para imprimir usando el formato nativo de Polars
        :param df: DataFrame de Polars a imprimir
        :param row_print: Numero de filas a imprimir
        :return: None
        """
        try:
            df_to_print = df.head(row_print)

            # Configurar opciones de display de Polars con UTF8_FULL
            with pl.Config(
                    tbl_rows=row_print,
                    tbl_cols=-1,  # Mostrar todas las columnas
                    tbl_width_chars=65000,  # Ancho máximo
                    fmt_str_lengths=1000,  # Longitud máxima de strings
                    tbl_formatting="UTF8_FULL"  # Formato UTF8 completo con bordes bonitos
            ):
                print(df_to_print)

        except Exception as e:
            print(f"Error al imprimir con formato nativo de Polars: {e}")
            print(df.head(row_print))

    def print_tabulate_df(self, df: pd.DataFrame | pl.DataFrame, show_index: bool = False, headers: list | None = None, row_print: int = 10, table_fmt: str = "fancy_grid"):
        """
        Metodo para imprimir en consola un dataframe tabulado
        :param table_fmt:
        :param df: Dataframe a imprimir
        :param show_index: Mostrar indices de fila
        :param headers: Columnas del dataframe (Si es None los muestra todos)
        :param row_print: Numero de filas a imprimir
        :return: None
        """
        try:

            # -- En caso de que sea un polars df, llamamos al metodo de polars, si es de pandas, sigue el flujo
            if isinstance(df, pl.DataFrame):
                self.print_polars_native(df, row_print=row_print)
                return

            # Guardar la configuración actual
            old_options = pd.get_option("display.float_format")

            # Configurar pandas para desactivar la notación científica
            pd.set_option('display.float_format', '{:.2f}'.format)
            pd.set_option('display.max_rows', None)  # Mostrar todas las filas
            pd.set_option('display.max_columns', None)  # Mostrar todas las columnas
            pd.set_option('display.width', None)  # Ancho de la consola sin truncamiento
            pd.set_option('display.max_colwidth', None)  # Mostrar toda la longitud de las cadenas en cada columna

            # Convertir el dataframe a string
            df = df.astype(str)

            if headers is not None:
                table = tabulate(df.head(row_print), showindex=show_index, headers=headers, tablefmt=table_fmt)
                table = table.replace('<th>', '<th style="background-color: lightgrey;">')
                print(table)
            else:
                table = tabulate(df.head(row_print), showindex=show_index, headers=df.columns, tablefmt=table_fmt)
                table = table.replace('<th>', '<th style="background-color: lightgrey;">')
                print(table)

            # Resetear las opciones de pandas a la configuración original
            pd.set_option("display.float_format", old_options)

            return
        except ValueError:
            print("NO SE PUEDE PINTAR TABULATE, PONGO EL PRINT")
            print(df.head(row_print))

    def header_print(self, text_to_print: str, text_color="light_yellow", bg_color=None, style="bright"):
        if self.show_info:
            stringVar = f'\n############################################################################################################################\n' \
                        f'############################################################################################################################\n' \
                        f'-------------------------------    {text_to_print}\n' \
                        f'############################################################################################################################\n' \
                        f'############################################################################################################################\n'

            print(self.print_colorama(stringVar, text_color=text_color, bg_color=bg_color, style=style))

    def intro_print(self, text_to_print: str, text_color="light_cyan", bg_color=None, style="bright"):
        if self.show_info:
            stringVar = f'\n#######################################################################################\n' \
                  f'----    {text_to_print}\n' \
                  f'#######################################################################################\n'

            print(self.print_colorama(stringVar, text_color=text_color, bg_color=bg_color, style=style))

    def sub_intro_print(self, text_to_print: str, text_color="light_white", bg_color=None, style=None):
        if self.show_info:
            print(f"\n{self.print_colorama('---> ' + text_to_print, text_color, bg_color, style)}")

    def info_print(self, text_to_print: str, text_color="light_green", bg_color=None, style="bright"):
        if self.show_info:
            print(f"{self.print_colorama('--------> [INFO]: ' + text_to_print, text_color, bg_color, style)}")

    def warning_print(self, text_to_print: str, text_color="light_white", bg_color="light_red", style="bright", category=UserWarning, stacklevel=2):
        if self.show_info:
            print(f"{self.print_colorama('--------> [WARNING]: ' + text_to_print, text_color, bg_color, style)}")
            warnings.warn(f"------> [WARNING]: {text_to_print}", category=category, stacklevel=stacklevel)

    @staticmethod
    def print_colorama(text, text_color=None, bg_color=None, style=None):
        """
        Imprime texto en consola con el color y estilo especificado.

        Args:
            text (str): El texto a imprimir.
            text_color (str): El color del texto. Debe ser uno de los siguientes:
                'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', None (por defecto).
            bg_color (str): El color del fondo. Debe ser uno de los siguientes:
                'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', None (por defecto).
            style (str): El estilo del texto. Debe ser uno de los siguientes:
                'reset', 'bright', 'dim', 'normal', None (por defecto).

        Raises:
            ValueError: Si el color o estilo especificado no es válido.
        """
        colors = {
            'black': Fore.BLACK,
            'red': Fore.RED,
            'green': Fore.GREEN,
            'yellow': Fore.YELLOW,
            'blue': Fore.BLUE,
            'magenta': Fore.MAGENTA,
            'cyan': Fore.CYAN,
            'white': Fore.WHITE,

            'light_black': Fore.LIGHTBLACK_EX,
            'light_red': Fore.LIGHTRED_EX,
            'light_green': Fore.LIGHTGREEN_EX,
            'light_yellow': Fore.LIGHTYELLOW_EX,
            'light_blue': Fore.LIGHTBLUE_EX,
            'light_magenta': Fore.LIGHTMAGENTA_EX,
            'light_cyan': Fore.LIGHTCYAN_EX,
            'light_white': Fore.LIGHTWHITE_EX,
            None: ''
        }

        bg_colors = {
            'black': Back.BLACK,
            'red': Back.RED,
            'green': Back.GREEN,
            'yellow': Back.YELLOW,
            'blue': Back.BLUE,
            'magenta': Back.MAGENTA,
            'cyan': Back.CYAN,
            'white': Back.WHITE,

            'light_black': Back.LIGHTBLACK_EX,
            'light_red': Back.LIGHTRED_EX,
            'light_green': Back.LIGHTGREEN_EX,
            'light_yellow': Back.LIGHTYELLOW_EX,
            'light_blue': Back.LIGHTBLUE_EX,
            'light_magenta': Back.LIGHTMAGENTA_EX,
            'light_cyan': Back.LIGHTCYAN_EX,
            'light_white': Back.LIGHTWHITE_EX,
            None: ''
        }

        styles = {
            'reset': Style.RESET_ALL,
            'bright': Style.BRIGHT,
            'dim': Style.DIM,
            'normal': Style.NORMAL,
            None: ''
        }

        if text_color not in colors:
            raise ValueError("Color de texto no válido. Los colores válidos son: black, red, green, yellow, blue, magenta, cyan, white, None")

        if bg_color not in bg_colors:
            raise ValueError("Color de fondo no válido. Los colores válidos son: black, red, green, yellow, blue, magenta, cyan, white, None")

        if style not in styles:
            raise ValueError("Estilo de texto no válido. Los estilos válidos son: reset, bright, dim, normal, None")

        color_seq = colors[text_color] + bg_colors[bg_color] + styles[style]
        reset_seq = Style.RESET_ALL

        return color_seq + text + reset_seq



from info_tools import InfoTools
from pandas_tools import PandasTools
from polars_tools import PolarsTools
from os_tools import OsTools
from singleton_tools import SingletonMeta
import datetime
import functools
import inspect
import polars as pl
import pandas as pd


class ConstantsAndTools(metaclass=SingletonMeta):
    def __init__(self):
        self.IT: InfoTools = InfoTools()
        self.PdT: PandasTools = PandasTools()
        self.PlT: PolarsTools = PolarsTools()
        self.OT: OsTools = OsTools()

        # -- Creacion de paths
        self.data_path, self.input_path, self.output_path = self.create_data_directories()

    def create_data_directories(self, root_path: str = "./") -> tuple:
        """
        Metodo que crea la carpeta data con subcarpetas input_data y output_data y retorna las 3 rutas
        :param root_path:
        :return:
        """

        # -- Carpeta data
        self.OT.create_folder_if_not_exists(f"{root_path}data")

        # -- Carpeta input_data
        self.OT.create_folder_if_not_exists(f"{root_path}data/input_data")

        # -- Carpeta output_data
        self.OT.create_folder_if_not_exists(f"{root_path}data/output_data")

        return f"{root_path}data", f"{root_path}data/input_data", f"{root_path}data/output_data"

    @staticmethod
    def hx_info_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            IT: InfoTools = InfoTools()
            start_time = datetime.datetime.now()
            IT.intro_print(f"[INFO DECORATOR] Ejecutando función / método: {func.__name__}")
            IT.info_print(f"🕒 Inicio: {start_time}", "light_magenta")

            sig = inspect.signature(func)
            bound_args = sig.bind(*args, **kwargs)
            bound_args.apply_defaults()

            IT.sub_intro_print("Parámetros:")
            for name, value in bound_args.arguments.items():
                IT.info_print(f"{name} = {value}")

            try:
                result = func(*args, **kwargs)
                if isinstance(result, pl.DataFrame) or isinstance(result, pd.DataFrame):
                    IT.info_print(f"No se pinta el resultado porque es un dataframe")
                else:
                    IT.info_print(f"✅ Resultado: {result}" if result is not None else "ℹ️ La función no devuelve nada")
            except Exception as e:
                IT.warning_print(f"❌ Error durante la ejecución: {e}")
                raise

            end_time = datetime.datetime.now()
            duration = end_time - start_time
            IT.sub_intro_print("Ejecucion correcta")
            IT.info_print(f"🕒 Tiempo de ejecución: {duration.total_seconds():.6f} segundos", "light_magenta")
            return result

        return wrapper


from setuptools import setup, find_packages

setup(
    name="constants_and_tools",
    version="2.2.1",
    author="Aletheia_corp",
    author_email="dsarabiatorres@gmail.com",
    description="Librería para desarrollo rápido y debug",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/aletheIA-Corp/constants_and_tools",
    packages=find_packages(),  # Busca automáticamente todos los paquetes
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "colorama==0.4.6",
        "openpyxl==3.1.5",
        "pandas==2.3.2",
        "pyarrow==21.0.0",
        "polars==1.33.1",
        "tabulate==0.9.0"
    ],
    python_requires=">=3.10",
)



import glob
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from singleton_tools import SingletonMeta
import os
from typing import Dict, List, Optional
from pathlib import Path


class OsTools(metaclass=SingletonMeta):
    def __init__(self):
        pass

    @staticmethod
    def execute_process_pool(methods: Dict) -> Dict:
        """
        Metodo que ejecuta el poolexecutor y retorna el diccionario con las respuestas
        Adecuado para tareas que consuman procesamiento CPU
        :param methods: EXAMPLE:
            - Con args: {'nombre_metodo_1': (mi_metodo, (param1,)), 'nombre_metodo_2': (mi_metodo2, (param1, param2,))}
            - Sin args: {'nombre_metodo_1': mi_metodo, 'nombre_metodo_2': mi_metodo2}
            - Mixto: {'nombre_metodo_1': mi_metodo, 'nombre_metodo_2': (mi_metodo2, (param1,))}
        NOTA: Los parámetros deben ir en formato tupla (param1, ). En caso de poner (param1) puede dar error
        :return: Diccionario con los resultados
        """

        # --------------------------------------------------------------------------
        # -- 1: Creo el diccionario que almacenará los resultados y arranco el Pool
        # --------------------------------------------------------------------------

        # ---- 1.1: Diccionario de resultados
        results: dict = {}

        # ---- 1.2: Entramos al Pool
        with ProcessPoolExecutor() as executor:

            # ------ 1.2.1: Diccionario que va a almacenar los futuros
            futures = {}

            # ------ 1.2.2: Iteramos por cada metodo del diccionario
            for name, value in methods.items():

                # -------- Detectar si value es una tupla con (func, args) o solo una función
                if isinstance(value, tuple) and len(value) == 2:
                    func, args = value
                    futures[name] = executor.submit(func, *args)
                else:
                    # -------- Es solo una función sin argumentos
                    func = value
                    futures[name] = executor.submit(func)

            for name, future in futures.items():
                try:
                    results[name] = future.result()  # Obtiene el resultado de cada proceso
                except Exception as e:
                    print(f"Error en {name}: {e}")

        return results

    @staticmethod
    def execute_thread_pool(methods: Dict) -> Dict:
        """
        Méthod que ejecuta el ThreadPoolExecutor y retorna el diccionario con las respuestas.
        Adecuado para tareas I/O-bound (como llamadas a API o lectura/escritura de archivos),
        :param methods: EXAMPLE:
            - Con args: {'nombre_metodo_1': (mi_metodo, (param1,)), 'nombre_metodo_2': (mi_metodo2, (param1, param2,))}
            - Sin args: {'nombre_metodo_1': mi_metodo, 'nombre_metodo_2': mi_metodo2}
            - Mixto: {'nombre_metodo_1': mi_metodo, 'nombre_metodo_2': (mi_metodo2, (param1,))}
        NOTA: Los parámetros deben ir en formato tupla (param1, ). En caso de poner (param1) puede dar error
        :return: Diccionario con los resultados
        """

        # --------------------------------------------------------------------------
        # -- 1: Creo el diccionario que almacenará los resultados y arranco el Pool
        # --------------------------------------------------------------------------

        # ---- 1.1: Diccionario de resultados
        results: dict = {}

        # ---- 1.2: Entramos al Pool
        with ThreadPoolExecutor() as executor:

            # ------ 1.2.1: Diccionario que va a almacenar los futuros
            futures = {}

            # ------ 1.2.2: Iteramos por cada metodo del diccionario
            for name, value in methods.items():

                # -------- Detectar si value es una tupla con (func, args) o solo una función
                if isinstance(value, tuple) and len(value) == 2:
                    func, args = value
                    futures[name] = executor.submit(func, *args)
                else:
                    # -------- Es solo una función sin argumentos
                    func = value
                    futures[name] = executor.submit(func)

            for name, future in futures.items():
                try:
                    results[name] = future.result()  # Obtiene el resultado de cada proceso
                except Exception as e:
                    print(f"Error en {name}: {e}")

        return results

    @staticmethod
    def create_folder_if_not_exists(folder_path: str) -> None:
        """
        Metodo para crear una carpeta si no existe ya
        :param folder_path:
        :return:
        """

        if not os.path.exists(folder_path):
            os.mkdir(folder_path)

    @staticmethod
    def get_path_files_by_extension(
            folder_path: str,
            extension: Optional[str] = None,
            substring: Optional[str] = None,
            recursive: bool = False,
            case_sensitive: bool = False
    ) -> List[str]:
        """
        Metodo genérico para obtener una lista de archivos de una carpeta según diversos criterios.

        :param folder_path: Ruta de la carpeta a buscar. EXAMPLE: 'data/input_data/extracted_files'
        :param extension: Extensión de archivo a filtrar (sin punto). EXAMPLE: 'png', 'pdf'. Si es None, busca todos.
        :param substring: Subcadena que debe contener el nombre del archivo. EXAMPLE: 'imagen', '2024'. Si es None, no filtra.
        :param recursive: Si True, busca en subcarpetas recursivamente. Default: False
        :param case_sensitive: Si True, la búsqueda distingue mayúsculas/minúsculas. Default: False
        :return: Lista de rutas de archivos que cumplen los criterios. EXAMPLE: ['img1.png', 'img2.png']
        """

        # -- 1: Construyo el patrón de búsqueda, si es recursivo, se va a buscar en las anidadas
        if recursive:
            pattern = f"{folder_path}/**/*"
        else:
            pattern = f"{folder_path}/*"

        # -- 2: Configuro la extension (si ya viene con punto lo elimino)
        if extension:
            pattern += f".{extension.replace('.', '')}"

        # -- 3: Obtengo todos los archivos que coinciden
        files = glob.glob(pattern, recursive=recursive)

        # -- 4: Filtro por substring y retorno
        if substring:
            if case_sensitive:
                files = [f for f in files if substring in Path(f).name]
            else:
                substring_lower = substring.lower()
                files = [f for f in files if substring_lower in Path(f).name.lower()]

        return files
